;--------------------------------------------------------------------
;                   FRONTEND PROGRAM FOR TRANSPHERE
;
; This IDL source code sets up all the necessary input files (with
; the exception of the frequency.inp, dustopac_1.inp and dustopac.inp
; files that need to be there already) for a problem of a spherical
; circumstellar cloud. It then runs the radiative transfer code, and
; reads in the data, and shows two figures: one for the T(R) and one
; for the SED.
;--------------------------------------------------------------------
@readopac.pro
@integrate.pro
@read.pro
@bplanck.pro
@natconst.pro
;
; Main setup parameters
;
rstar    = 5.3453d0 * RS  ; Stellar radius
mstar    = 1 * MS         ; Stellar mass
tstar    = 100.          ; Stellar temperature
nr       = 200            ; Nr of radial shells for grid
nref     = 100            ; Nr of refinement points
rin      = 20. * AU       ; Inner radius of shell
rout     = 1.2d5 * AU     ; Outer radius of shell
rref     = 500. * AU      ; Refinement radius
r0       = 1d3 * AU       ; Reference radius
rho0     = 3.46d-18        ; Density at reference radius
plrhoin  = -1.8           ; Powerlaw for rho
plrhoout = -10            ; Powerlaw for rho
nriter   = 30             ; Maximum nr of iterations
convcrit = 0.00001        ; Convergence criterion
init     = 1              ; Yes, we start with a good initial guess
ncst     = 10             ; Nr of rays for star
ncex     = 30             ; Nr of rays between star and Rin
ncnr     = 1              ; Nr of rays per radial grid point
itypemw  = 1              ; Type of mu weighting
idump    = 1              ; Dump convergence history
localdust= 0              ; Dust opacity local?
tbg      = 1d3            ; Background temperature
dilute   = 1d-8          ; Dilution of background intensity
;
; Read the silicate opacity
;
o   = readopac(nr='sil')
nf  = n_elements(o.freq)
;
; Write the actual opacity as a function of R
;
if keyword_set(localdust) then begin
   spawn,'rm -f dustopac_1.inp'
   openw,1,'dustopac.inp'
   printf,1,'1               Format number of this file'
   printf,1,'1               Nr of dust species'
   printf,1,'============================================================================'
   printf,1,'-1              Way in which this dust species is read (-1=file)'
   printf,1,'0               Extension of name of dustopac_***.inp file'
   printf,1,'----------------------------------------------------------------------------'
   close,1
   openw,1,'dustopac_0.inp'
   printf,1,nr
   printf,1,nf,1
   printf,1,' '
   redux=1.d0
   for ir=0,nr-1 do begin
;   if ir lt nr/2 then redux=0.1 else redux=1.d0
      for inu=0,nf-1 do printf,1,o.cabs[inu]*redux
      for inu=0,nf-1 do printf,1,o.csca[inu]*redux
      printf,1,' '
   endfor
   close,1
endif else begin
   spawn,'rm -f dustopac_0.inp'
   openw,1,'dustopac.inp'
   printf,1,'1               Format number of this file'
   printf,1,'1               Nr of dust species'
   printf,1,'============================================================================'
   printf,1,'-1              Way in which this dust species is read (-1=file)'
   printf,1,'1               Extension of name of dustopac_***.inp file'
   printf,1,'----------------------------------------------------------------------------'
   close,1
   openw,1,'dustopac_1.inp'
   printf,1,nf,1
   printf,1,' '
   for inu=0,nf-1 do printf,1,o.cabs[inu]
   for inu=0,nf-1 do printf,1,o.csca[inu]
   close,1
endelse
;
; Find the kappa at 550 nm
;
if keyword_set(localdust) then begin
   openr,1,'dustopac_0.inp'
   nnr=0
   nnf=0
   idum=0
   readf,1,nnr
   readf,1,nnf,idum
   if nnr ne nr then stop
   if nnf ne nf then stop
   kap=dblarr(nf,2,nr)
   readf,1,kap
   close,1
   ivis   = 0
   fvis   = 1d4 * cc / 0.55d0 
   for inu=1,nf-1 do if((o.freq[inu]-fvis)*   $
                        (o.freq[inu-1]-fvis) lt 0.d0) then ivis=inu
   eps    = (fvis-o.freq[ivis-1]) / (o.freq[ivis]-o.freq[ivis-1])
   kappa  = transpose((1.d0-eps)*kap[ivis-1,0,*]+eps*kap[ivis,0,*],[2,0,1])
endif else begin
   openr,1,'dustopac_1.inp'
   nnf=0
   idum=0
   readf,1,nnf,idum
   if nnf ne nf then stop
   kap=dblarr(nf,2)
   readf,1,kap
   close,1
   ivis   = 0
   fvis   = 1d4 * cc / 0.55d0 
   for inu=1,nf-1 do if((o.freq[inu]-fvis)*   $
                        (o.freq[inu-1]-fvis) lt 0.d0) then ivis=inu
   eps    = (fvis-o.freq[ivis-1]) / (o.freq[ivis]-o.freq[ivis-1])
   kappa  = (1.d0-eps)*kap[ivis-1,0]+eps*kap[ivis,0]
endelse
;
; Make the R grid
;
if nref eq 0 then begin
   ;
   ; Simple log grid
   ;
   r             = rin * (rout/rin)^(dindgen(nr)/(nr-1.d0))
endif else begin
   ;
   ; Log grid with refinement at inner edge
   ;
   lgr0          = alog10(rin)
   lgr1          = alog10(rout)
   lgrr          = alog10(rref)
   n1            = nref
   n2            = nr-nref-1
   n             = dindgen(nr)-n1
   c             = (lgr1-lgrr)/(1.d0*n2)
   b             = c/(lgrr-lgr0)
   r             = dblarr(nr)
   r[0:n1]       = (lgrr-lgr0)*exp(b*n[0:n1])
   r[n1+1:n1+n2] = (lgrr-lgr0)+c*n[n1+1:n1+n2]
   r             = r-(r[0]+(r[0]-r[1]))
   r             = r*(lgr1-lgr0)/(max(r))+lgr0
   r             = 10^r
endelse
;
; Make the density profile such that the optical depth is correct
;
ii     = where(r le r0)
rhodust    = dblarr(nr)
rhodust[ii]  = 1d-2*rho0 * (r[ii]/r0)^(plrhoin)
ii     = where(r gt r0)
rhodust[ii]  = 1d-2*rho0 * (r[ii]/r0)^(plrhoout)
tau    = integrate(r,rhodust*kappa)
print,'Tau_V             = ',tau
mass   = 1d2*integrate(r,4*!pi*r^2*rhodust)
print,'Mass (dust+gas)   = ',mass/MS,' Msun'
;
; Make the input file for the transphere program
;
openw,1,'transphere.inp'
printf,1,2
printf,1,nriter  
printf,1,convcrit
printf,1,ncst
printf,1,ncex
printf,1,ncnr
printf,1,itypemw
printf,1,idump
printf,1,init
close,1
;
; Make the stellar information file
; (mstar and tstar are irrelevant; they are there for historical reasons)
;
openw,1,'starinfo.inp'
printf,1,1
printf,1,rstar
printf,1,mstar
printf,1,tstar
close,1
;
; The stellar spectrum
;
openw,1,'starspectrum.inp'
printf,1,nf
for inu=0,nf-1 do begin
   printf,1,o.freq[inu],(rstar/pc)^2*!dpi*bplanck(o.freq[inu],tstar)
endfor
close,1
;
; The exterior spectrum
;
openw,1,'external_meanint.inp'
printf,1,nf
for inu=0,nf-1 do begin
   printf,1,o.freq[inu],dilute*bplanck(o.freq[inu],tbg)
endfor
close,1
;
; Write the envelope structure
;
openw,1,'envstruct.inp'
printf,1,nr
printf,1,' '
for ir=0,nr-1 do begin
   printf,1,r[ir],rhodust[ir],30.d0,format='(3(E13.6,1X))'
endfor
close,1
;
; Call the transphere routine
;
spawn,'../src/transphere'
;
; Read the output
;
s=read_spectrum()
a=read_convhist()
zz=size(a.temp)
ni=zz(2)
;
; Plot the output to postscript files
;
thick=3.0      ; Postscript line thickness
size=1.4       ; Postscript char size
!p.thick=thick
!x.thick=thick
!y.thick=thick
!z.thick=thick
!p.charthick=thick
!p.charsize=size
;
; ... The spectrum
;
set_plot,'ps'
device,file='spectrum.ps'
plot_spectrum,s,/lsun,yrange=[1d-6,100],xrange=[1,1e4]
device,/close
set_plot,'x'
;
; ... The dust temperature
;
set_plot,'ps'
device,file='dusttemp.ps'
plot,a.r/AU,a.temp(*,ni-1),/xlog,/ylog,xtitle='R [AU]',ytitle='T [K]',$
    title='Dust temperature (dotted = previous iterations)'
for i=0,ni-2 do oplot,a.r/AU,a.temp(*,i),line=1
device,/close
set_plot,'x'
;
; Ghostview them 
;
spawn,'gv spectrum.ps &'
spawn,'gv dusttemp.ps &'
;
!p.thick=1
!x.thick=1
!y.thick=1
!z.thick=1
!p.charthick=1
!p.charsize=1
;
end
