c     ==============================================================
c
c                         >>>> TRANSPHERE <<<<
c
c     A DUST CONTINUUM RADIATIVE TRANSFER CODE FOR SPHERICAL SYMMETRY
c
c              written by C.P. Dullemond (C) March 2002
c
c     This code solves the radiative transfer problems in spherical
c     symmetry of a dust envelope surrounding a star with a certain
c     temperature and luminosity.  The dust grains are assumed to be in
c     radiative equilibrium, and the total flux is conserved. The code
c     solves the dust temperature at every radius and returns the
c     spectral energy distribution of the system. The code can solve
c     problems of any optical depth within about 10 iterations
c     only. This powerful capability is made possible through the use of
c     variable eddington factors and variable mean opacities.
c     
c     In the present (downscaled) form, the code can handle only a
c     single dust species and size, and dust scattering is not
c     included. Also, the code only reads in the stellar temperature and
c     size, and assumes the star to be a blackbody. The full version of
c     the code does not have these limitations and can input a realistic
c     spectrum, can handle multiple species and can include scattering
c     (in the isotropic limit).
c     
c     If you like the present downscaled version of the code, and wish
c     to use it for scientific use, you can obtain a full version of the
c     code by contacting the author via email:
c     dullemon@mpa-garching.mpg.de. Conditions for use can then be
c     agreed upon through email exchange.
c     
c                  ----------------------------------
c                  !!!!!!!  Copyright notice: !!!!!!!
c                  ----------------------------------
c     
c     The present code is public domain, but remains under copyright.
c     The code or any part of the code) may NOT be used for (any part
c     of) a publication in a refereed scientific journal without written
c     permission from the author. You can reach the author via email:
c     dullemon@mpa-garching.mpg.de.
c     
c     This package may be distributed to others, under the condition
c     that it it kept intact and complete, including this README file.
c     
c     Note: 
c     
c       The package contains a few subroutines from Numerical Recipes. 
c       This is copyrighted software, so note the copyright restrictions 
c       of that software. For more information on Numerical Recipes, see 
c       their webpage: 
c           www.ulib.org/webRoot/Books/Numerical_Recipes/bookf.html
c     
c     Note:
c     
c       The program TRANSPHERE is written in fortran 77. An example setup 
c       is created by a program written in the IDL programming language 
c       (see www.rsinc.com). Also the example read and plot routines for 
c       the output are written in IDL. IDL is licenced software, and can 
c       be purchased from Research Systems Inc. 
c     
c     Website:
c     
c        More information about various astrophysical software packages 
c        written by C.P. Dullemond can be found on the following homepage:
c     
c            http://www.mpa-garching.mpg.de/~dullemon
c     
c            (NOTE: This page remains here until 2004; After leaving MPA,
c             this page will be moved. Unfortunately, MPA will not support
c             a forward link to my new address.)
c     
c     ==============================================================

#include "main.h"

#define PLANCKTAB_NT 100
#define PLANCKTAB_T0 (1.D-1)
#define PLANCKTAB_T1 (1.D+4)

c     --------------------------------------------------------------
c                             MAIN ROUTINE
c     --------------------------------------------------------------
      program main
      implicit none
c
      call read_all_data()
      call do_all_radtrans()
      call write_temperatures()
      call write_aux_data()
      call write_spectrum()
c
      end



c     --------------------------------------------------------------
c                         READ ALL THE DATA IN
c     --------------------------------------------------------------
      subroutine read_all_data()
      implicit none
c
#include "common_grid.h"
#include "common_dust.h"
#include "common_radtrans.h"
c
      doubleprecision rstar,a,b,c,d,e,dum
      integer nr,ir,nrp,nf,inu,nnf
      integer iformat
      logical f_ex,fexf,fexo,fex
c
c     Read the fullvrt input file
c
      global_init=0
      global_dump_intens=0
      global_imethod=1
      global_isrf_for_imgspec=0
      open(unit=1,file='transphere.inp',status='old')
      read(1,*) iformat
      read(1,*) vet_itermax
      read(1,*) vet_convcrit
      read(1,*) global_ncst
      read(1,*) global_ncex
      read(1,*) global_ncnr
      if(iformat.ge.2) then
          read(1,*) global_itypemw
      else
          global_itypemw = 1
      endif
      read(1,*) global_idump
      read(1,*,end=100,err=100) global_init
      read(1,*,end=100,err=100) global_dump_intens
      read(1,*,end=100,err=100) global_imethod
      read(1,*,end=100,err=100) global_isrf_for_imgspec
      close(1)
 100  continue
c
c     Make note:
c
      if(global_isrf_for_imgspec.eq.0) then
          write(*,*) 'Note: In the output spectrum the ISRF is '
          write(*,*) '      NOT included. You can include it by '
          write(*,*) '      setting the global_isrf_for_imgspec=1 '
          write(*,*) '      (see transphere.inp). In the RT iteration '
          write(*,*) '      it IS always included.'
      else
          write(*,*) 'Note: In the output spectrum the ISRF is '
          write(*,*) '      included. You can EXclude it by '
          write(*,*) '      setting the global_isrf_for_imgspec=0 '
          write(*,*) '      (see transphere.inp)'
      endif
c
c     Read the primary temperature and the density
c
      open(unit=1,file='envstruct.inp',status='old')
      read(1,*) nr
      if(nr.gt.FRSIZE_R) then
          write(*,*) 'ERROR:'
          write(*,*) 'file envstruct.inp: R grid too large'
          stop 13
      endif
      grid_nr = nr
      do ir=1,nr
          read(1,*) a,b,c
          grid_r(ir)=a
          srt_rho(ir)=b
          srt_temp(ir)=c
          if(srt_temp(ir).le.0.d0) srt_temp(ir)=1.d0
      enddo
      close(1)
c
c     Read the frequency array and the opacities
c
      inquire(file='frequency.inp',exist=fexf)
      inquire(file='dustopac_1.inp',exist=fexo)
      if((.not.fexf).or.(.not.fexo)) then
          inquire(file='dustopac_0.inp',exist=fexo)
          if(.not.fexo) then 
              write(*,*) 'Need both frequency.inp and dustopac_1.inp'
              stop 13
          endif
      endif
      open(unit=1,file='frequency.inp',status='old')
      read(1,*) nf
      if(nf.gt.FRSIZE_FREQ) then
          write(*,*) 'ERROR in main:'
          write(*,*) 'Frequency grid too large'
          stop 13
      endif
      do inu=1,nf
          read(1,*) a
          freq_nu(inu) = a
      enddo
      dust_setup_nrspecies=1
      dust_setup_nrsizes(1)=1
      freq_nr = nf
      call read_dustdata(grid_nr)
c
c     Read the star information
c
      open(unit=1,file='starinfo.inp',status='old')
      read(1,*) iformat
      read(1,*) star_r
      read(1,*) star_m
      read(1,*) star_t
      close(1)
c
c     Read the star spectrum
c
      inquire(file='starspectrum.inp',exist=fex)
      if(.not.fex) then
          write(*,*) 'ERROR: In version 2 of TRANSPHERE you must '
          write(*,*) '       specify the stellar spectrum,'
          write(*,*) '       in the file starspectrum.inp.'
          write(*,*) '       This is the stellar flux (erg/s/cm^2/Hz)'
          write(*,*) '       as observed at 1 parsec distance.'
          stop
      endif
      open(unit=1,file='starspectrum.inp',status='old')
      read(1,*) nnf
      if(nnf.ne.freq_nr) then
          write(*,*) 'ERROR reading starspectrum.inp.'
          write(*,*) '      Nr of freq not equal to that of ',
     %               'frequency.inp'
          stop 
      endif
      do inu=1,freq_nr
          read(1,*) dum,star_int(inu)
          if(abs((dum-freq_nu(inu))/(dum+freq_nu(inu))).gt.1d-4) then
              write(*,*) 'ERROR while reading starspectrum.inp.'
              write(*,*) '      Frequency grid not consistent with'
              write(*,*) '      that of frequency.inp.'
              stop
          endif
      enddo
      close(1)
      do inu=1,freq_nr
          star_int(inu) = 3.0308410d36 * star_int(inu) / star_r**2
      enddo
c
c     Read the external mean intensity file
c
      inquire(file='external_meanint.inp',exist=fex)
      if(.not.fex) then
          write(*,*) 'ERROR: In version 2 of TRANSPHERE you must '
          write(*,*) '       specify J(nu) of the external environment,'
          write(*,*) '       in the file external_meanint.inp.'
          write(*,*) '       Put to zero, if you want same setup of '
          write(*,*) '       version 1.'
          stop
      endif
      open(unit=1,file='external_meanint.inp',status='old')
      read(1,*) nnf
      if(nnf.ne.freq_nr) then
          write(*,*) 'ERROR reading external_meanint.inp.'
          write(*,*) '      Nr of freq not equal to that of ',
     %               'frequency.inp'
          stop 
      endif
      do inu=1,freq_nr
          read(1,*) dum,ext_meanint(inu)
          if(abs((dum-freq_nu(inu))/(dum+freq_nu(inu))).gt.1d-4) then
              write(*,*) 'ERROR while reading external_meanint.inp.'
              write(*,*) '      Frequency grid not consistent with'
              write(*,*) '      that of frequency.inp.'
              stop
          endif
      enddo
      close(1)
c
      end



c     --------------------------------------------------------------
c                         PERFORM ALL RT
c     --------------------------------------------------------------
      subroutine do_all_radtrans()
      implicit none
c
#include "common_grid.h"
#include "common_dust.h"
#include "common_radtrans.h"
#include "common_transfer.h"
c
      doubleprecision dumtemp,find_dust_kappa,t0,t1
      integer ir,inu,nt
      parameter(nt=PLANCKTAB_NT)
      parameter(t0=PLANCKTAB_T0)
      parameter(t1=PLANCKTAB_T1)
c
c     First make a table of opacities, both the freq-dependent
c     ones, and the Planck-mean ones.
c
      if(dust_opacity_tempdep.ne.0) then
          write(*,*) 'Sorry, but fullrt cannot use temperature'
          write(*,*) 'dependent opacities for the moment...'
          stop 13
      endif
      dumtemp = 10.d0
      do ir=1,grid_nr
          do inu=1,freq_nr
              optab_kappa(inu,ir) = find_dust_kappa(inu,1,1,ir,t0,1,0)
          enddo
      enddo
      call compute_planckopac(freq_nr,nt,grid_nr,freq_nu,t0,t1,
     %                        optab_kappa)
c
c     Do RT and check energy conservation
c
      call do_radtrans(global_imethod)
c      call check_encons()
c
      end




c     --------------------------------------------------------------
c                DO THE TRANSFER IN RADIAL DIRECTION
c     --------------------------------------------------------------
      subroutine do_radtrans(imethod)
      implicit none
c
#include "common_grid.h"
#include "common_transfer.h"
#include "common_radtrans.h"
c
      doubleprecision bplanck
      integer ierror,ing,iali,imethod,intdump,ir,inu,intorder
c
c     Compute the source term for the transfer
c
      do ir=1,grid_nr
          srt_qfpi(ir) = 0.d0
      enddo
c
c     Now do the full transfer 
c
      ing = 0            
      intdump=0
      intorder=1
      call do_vet_transfer(grid_nr,freq_nr,global_ncst,global_ncex,
     %              global_ncnr,grid_r,freq_nu,srt_rho,srt_qfpi,
     %              star_r,star_int,vet_itermax,vet_convcrit,ing,
     %              intorder,global_itypemw,global_idump,intdump,
     %              imethod,
     %              global_init,ierror,optab_kappa,srt_temp,
     %              vet_jjme,vet_hhme,vet_jj,vet_hh,
     %              vet_jnu,vet_hnu,vet_fj,vet_fnu,vet_kapt,
     %              vet_kapj,vet_kaph,vet_corfact)
c
c     Check error message
c
      if(ierror.ne.0) then
          write(*,*) '===== PROBLEM IN FULLRT AT IR=',ir,' ====='
      endif
c
      end




c     --------------------------------------------------------------
c                THE BLACKBODY PLANCK FUNCTION B_nu(T)
c
c     This function computes the Blackbody function 
c
c                    2 h nu^3 / c^2
c        B_nu(T)  = -------------------    [ erg / cm^2 s ster Hz ]
c                   exp(-h nu / kT) - 1
c
c     ARGUMENTS:
c        nu    [Hz]            = Frequency
c        temp  [K]             = Electron temperature
c     --------------------------------------------------------------
      function bplanck(temp,nu)
      implicit none
      doubleprecision temp
      doubleprecision nu
      doubleprecision bplanck
c
      if(temp.eq.0.d0) then 
          bplanck = 0.d0
          return
      endif
c
      bplanck = 1.47455d-47 * nu * nu * nu / 
     %      (exp(4.7989d-11 * nu / temp)-1.d0) + 1.d-290
c
      return
      end





c     --------------------------------------------------------------
c                    WRITE THE MAIN OUTPUT FILE
c     --------------------------------------------------------------
      subroutine write_temperatures()
      implicit none
c
#include "common_grid.h"
#include "common_radtrans.h"
c
      integer ir
c
      open(unit=1,file='envstruct.dat',status='unknown')
      write(1,*) grid_nr
      write(1,*)
      do ir=1,grid_nr
          write(1,10) grid_r(ir),srt_rho(ir),srt_temp(ir)
 10       format(3(E13.6,1X))
      enddo
      close(1)
c
      end


c     --------------------------------------------------------------
c                  WRITE SOME OF THE AUXILIARY DATA
c     --------------------------------------------------------------
      subroutine write_aux_data()
      implicit none
c
#include "common_grid.h"
#include "common_radtrans.h"
c
      integer ir
c
      open(unit=1,file='transphere.dat',status='unknown')
      write(1,*) grid_nr
      write(1,*)
      do ir=1,grid_nr
          write(1,10) vet_hh(ir),vet_jj(ir),
     %                vet_fj(ir),srt_qfpi(ir)
 10       format(4(E13.6,1X))
      enddo
      close(1)
c
      end


