c     ==============================================================
c             MODULE FOR INTEGRATING THE MOMENT EQUATIONS 
c                     IN SPHERICAL COORDINATES
c
c     This module integrates the moment equations of radiative 
c     transfer in spherical coordinates, using flux-mean opacities.
c     In spherical transfer problems we are not faced with the
c     zero-flux problems encountered in passive disks (see paper
c     Dullemond, Van Zadelhoff & Natta 2002), so we can use the
c     flux-averaged opacity without problems.
c
c     ==============================================================

#include "main.h"


c     --------------------------------------------------------------
c                 DO INTEGRATION OF MOMENT EQUATIONS
c
c     This routine has the stellar luminosity as boundary condition.
c     For passive clouds there is no energy production within the
c     cloud, so that the luminosity is constant. But it is possible
c     to include energy sources or sinks. 
c
c     ARGUMENTS:
c       r          The R-grid
c       rho        The density
c       qfpi       The source term per unit gram per sterrad
c       l0         Boundary condition ir=1: the luminosity there
c       hj1        Boundary condition ir=nr: the ratio (H+Jext)/J there
c       fj         The eddington factor (f=0 means isotropic)
c       jext       The mean intensity of the external radiation field      
c       kaph       The H-mean opacity
c
c     OUTPUT:
c       jj         The mean intensity
c       hh         The Eddington flux
c     
c     NOTE: The f is chosen such that K = ( 1/3 + f ) J. 
c     --------------------------------------------------------------
      subroutine do_integr_momeq(nr,r,rho,qfpi,l0,hj1,fj,jext,kaph,
     %                           jj,hh,hhrt)
      implicit none
      integer nr
      doubleprecision r(FRSIZE_R),qfpi(FRSIZE_R),l0,hj1,jext
      doubleprecision rho(FRSIZE_R),fj(FRSIZE_R)
      doubleprecision kaph(FRSIZE_R)
      doubleprecision jj(FRSIZE_R),hh(FRSIZE_R),hhrt(FRSIZE_R)
c
      doubleprecision ll(FRSIZE_R),rk(FRSIZE_R),dum(FRSIZE_R),fpisq
      doubleprecision pi,rkorig
      parameter(pi  = 3.1415926535897932385d0)
      integer ir,iter
      logical error
c
c     Constant
c
      fpisq = (4.*pi)**2
c
c     The inner boundary condition:  L(R_0) = L
c
      ll(1) = l0
      hh(1) = ll(1) / ( fpisq*r(1)**2 )
c
c     Integrate L(z) from R_0 to R_out:
c
c      dL(R)
c      ----- = (4*pi)^2 R^2 q(R)
c       dR
c
c     ********** IS THERE A BUG HERE? (R instead of R^2) *********
c     ********** (NOT IMPORTANT, AS THERE IS NO SOURCE)  *********
c     ********** (WHILE IMPROVING... 22-04-04)           *********
c
      do ir=2,nr
          ll(ir) = ll(ir-1) + (4.*pi)**2 * 0.5d0 *
     %                 ( rho(ir) * qfpi(ir) * r(ir) + 
     %                   rho(ir-1) * qfpi(ir-1) * r(ir-1) ) *
     %                 ( r(ir) - r(ir-1) ) 
          hh(ir) = ll(ir) / ( fpisq*r(ir)**2 )
          if(hh(ir).lt.0.d0) stop 633
      enddo
c
c     The outer boundary condition: J = (H+Jext) / hj1
c     
      jj(nr)  = (abs(hh(nr))+jext) / hj1
c
c     Now compute R^2 K = ( 1/3 + f ) R^2 J here
c
      rk(nr)  = ( (1.d0/3.d0) + fj(nr) ) * jj(nr) * r(nr)**2
c
c     The first (outermost) value of the integrand
c
      dum(nr) = rk(nr) * (2.-3.*fj(nr)) / 
     %               ( (1.+3.*fj(nr)) * r(nr) ) -
     %              rho(nr) * kaph(nr) * 
     %              hh(nr) * r(nr)**2
c     
c     Now integrate K R^2 back to the inner boundary
c
c       d (R^2 K)   (2-3f) R^2 K                 L(R)
c       --------- - ------ ----- = - rho kappa -------
c          dR       (1+3f)   R                 (4pi)^2
c
c     (Sorry for the stupid "implicit integration". I did it since it
c     is very simple, and this part of the program is not time critical.)
c
      error=.false.
      do ir=nr-1,1,-1
          rkorig  = 0.d0
          rk(ir)  = rk(ir+1) - dum(ir+1) * ( r(ir+1) - r(ir) )
          do iter=1,20 
              dum(ir) = rk(ir) * (2.-3.*fj(ir)) / 
     %                  ( (1.+3.*fj(ir)) * r(ir) ) -
     %                  rho(ir) * kaph(ir) * 
     %                  hh(ir) * r(ir)**2
              rk(ir)  = rk(ir+1) - 0.5d0 * 
     %                  ( dum(ir) + dum(ir+1) ) *
     %                  ( r(ir+1) - r(ir) )
              if(abs(rk(ir)/rkorig-1.d0).lt.1.d-4) goto 30
              rkorig = rk(ir)
          enddo
          write(*,*) 'WEIRD ERROR: No convergence in integr iteration.'
          write(*,*) ir,rkorig,rk(ir),rk(ir+1),dum(ir),dum(ir+1)
          write(*,*) rho(ir),kaph(ir),ll(ir)
          stop 13
 30       continue
          jj(ir)  = rk(ir) / 
     %        ( ( (1.d0/3.d0) + fj(ir) ) * r(ir)**2 )
c
c         PROBLEM FIX (BUGFIX) 23.03.06: 
c         If jj drops below L/((4*pi)^2 r^2) then evidently something 
c         must have gone wrong, hopefully temporarily. So we correct
c         for this here
c
          if(jj(ir).lt.hh(ir)) then 
              error=.true.
c              write(*,*) 'Fixing low J...'
c              jj(ir) = hh(ir)
          endif
      enddo
      if(error) then
          open(unit=2,file='tt.dat')
          do ir=1,nr
              write(2,*) r(ir),jj(ir),hh(ir),fj(ir),rk(ir)
          enddo
          stop
      endif
c     
      end




c     --------------------------------------------------------------
c                 DO INTEGRATION OF MOMENT EQUATIONS
c               USING THE MORE STABLE METHOD WITH H_NU
c                              18.04.07
c
c     This is like the above routine, but now we use the hnu that
c     arises from the raytracing as the flux to use for the integration
c     of the jj(ir), but we normalize hnu on the basis of the hh(ir)
c     that follows from the normal moment equations. This method avoids
c     the use of the H-mean opacity, which goes berserk when H~=0.
c
c     ARGUMENTS:
c       r          The R-grid
c       rho        The density
c       qfpi       The source term per unit gram per sterrad
c       l0         Boundary condition ir=1: the luminosity there
c       hj1        Boundary condition ir=nr: the ratio (H+Jext)/J there
c       fj         The eddington factor (f=0 means isotropic)
c       jext       The mean intensity of the external radiation field      
c       kapnu      The kappa_nu
c
c     OUTPUT:
c       jj         The mean intensity
c       hh         The Eddington flux
c     
c     NOTE: The f is chosen such that K = ( 1/3 + f ) J. 
c     --------------------------------------------------------------
      subroutine do_integr_momnu(nr,nf,r,nu,rho,temp,qfpi,
     %                           l0,hj1,fj,jext,
     %                           hnu,alphanu,jj,hh,corfact)
      implicit none
      integer nr,nf
      doubleprecision r(FRSIZE_R),qfpi(FRSIZE_R),l0,hj1,jext
      doubleprecision rho(FRSIZE_R),fj(FRSIZE_R),temp(FRSIZE_R)
      doubleprecision hnu(FRSIZE_FREQ,FRSIZE_R),nu(FRSIZE_FREQ)
      doubleprecision alphanu(FRSIZE_FREQ,FRSIZE_R)
      doubleprecision jj(FRSIZE_R),hh(FRSIZE_R)
      doubleprecision corfact(FRSIZE_R),hkint(FRSIZE_R)
c
      doubleprecision ll(FRSIZE_R),rk(FRSIZE_R),dum(FRSIZE_R),fpisq
      doubleprecision pi,rkorig
      parameter(pi  = 3.1415926535897932385d0)
      integer ir,iter,inu
      logical error
c
c     Constant
c
      fpisq = (4.*pi)**2
c
c     The inner boundary condition:  L(R_0) = L
c
      ll(1) = l0
      hh(1) = ll(1) / ( fpisq*r(1)**2 )
c
c     Integrate L(z) from R_0 to R_out:
c
c      dL(R)
c      ----- = (4*pi)^2 R^2 q(R)
c       dR
c
c     ********** IS THERE A BUG HERE? (R instead of R^2) *********
c     ********** (NOT IMPORTANT, AS THERE IS NO SOURCE)  *********
c     ********** (WHILE IMPROVING... 22-04-04)           *********
c
c      do ir=2,nr
c          ll(ir) = ll(ir-1) + (4.*pi)**2 * 0.5d0 *
c     %                 ( rho(ir) * qfpi(ir) * r(ir) + 
c     %                   rho(ir-1) * qfpi(ir-1) * r(ir-1) ) *
c     %                 ( r(ir) - r(ir-1) ) 
c          hh(ir) = ll(ir) / ( fpisq*r(ir)**2 )
c          if(hh(ir).lt.0.d0) stop 633
c      enddo
c
c     ********** (FIXED 19-04-07)                        *********
c
      do ir=2,nr
          ll(ir) = ll(ir-1) + (4.*pi)**2 * 0.5d0 *
     %                 ( rho(ir) * qfpi(ir) * r(ir)**2 + 
     %                   rho(ir-1) * qfpi(ir-1) * r(ir-1)**2 ) *
     %                 ( r(ir) - r(ir-1) ) 
          hh(ir) = ll(ir) / ( fpisq*r(ir)**2 )
          if(hh(ir).lt.0.d0) stop 633
      enddo
c
c     The outer boundary condition: J = ( H + J_ext ) / hj1
c     
      jj(nr)  = (abs(hh(nr))+jext) / hj1
c
c     Now compute R^2 K = ( 1/3 + f ) R^2 J here
c
      rk(nr)  = ( (1.d0/3.d0) + fj(nr) ) * jj(nr) * r(nr)**2
c
c     For numerical stability it is important to 'correct' the
c     flux at every z, so that the freq-integrated flux is equal to 
c     the freq-integrated flux computed from the moment equations
c     (i.e. the integral of hnu over nu should equal hh).
c     
      do ir=1,nr
          call correct_flux(nf,nu,hnu(1,ir),hh(ir),
     %                 alphanu(1,ir),temp(ir),corfact(ir))
      enddo
c
c     Compute the integral H_nu * kappa_nu dnu at all z
c     
      do ir=1,nr
          hkint(ir) = 0.d0
          do inu=2,nf
              hkint(ir) = hkint(ir) + 0.5 * 
     %          ( hnu(inu,ir)*alphanu(inu,ir) + 
     %            hnu(inu-1,ir)*alphanu(inu-1,ir) ) *
     %          abs( nu(inu) - nu(inu-1) )
          enddo
      enddo
c
c     The first (outermost) value of the integrand
c
      dum(nr) = rk(nr) * (2.-3.*fj(nr)) / 
     %               ( (1.+3.*fj(nr)) * r(nr) ) -
     %              hkint(nr) * r(nr)**2
c     
c     Now integrate K R^2 back to the inner boundary
c
c       d (R^2 K)   (2-3f) R^2 K                 L(R)
c       --------- - ------ ----- = - rho kappa -------
c          dR       (1+3f)   R                 (4pi)^2
c
c     (Sorry for the stupid "implicit integration". I did it since it
c     is very simple, and this part of the program is not time critical.)
c
      error=.false.
      do ir=nr-1,1,-1
          rkorig  = 0.d0
          rk(ir)  = rk(ir+1) - dum(ir+1) * ( r(ir+1) - r(ir) )
          do iter=1,20 
              dum(ir) = rk(ir) * (2.-3.*fj(ir)) / 
     %                  ( (1.+3.*fj(ir)) * r(ir) ) -
     %                  hkint(ir) * r(ir)**2
              rk(ir)  = rk(ir+1) - 0.5d0 * 
     %                  ( dum(ir) + dum(ir+1) ) *
     %                  ( r(ir+1) - r(ir) )
              if(abs(rk(ir)/rkorig-1.d0).lt.1.d-4) goto 30
              rkorig = rk(ir)
          enddo
          write(*,*) 'WEIRD ERROR: No convergence in integr iteration.'
          write(*,*) ir,rkorig,rk(ir),rk(ir+1),dum(ir),dum(ir+1)
          write(*,*) hkint(ir),ll(ir)
          stop 13
 30       continue
          jj(ir)  = rk(ir) / 
     %        ( ( (1.d0/3.d0) + fj(ir) ) * r(ir)**2 )
c
c         PROBLEM FIX (BUGFIX) 23.03.06: 
c         If jj drops below L/((4*pi)^2 r^2) then evidently something 
c         must have gone wrong, hopefully temporarily. So we correct
c         for this here
c
c          if(jj(ir).lt.hh(ir)) then 
c              error=.true.
cc              write(*,*) 'Fixing low J...'
cc              jj(ir) = hh(ir)
c          endif
      enddo
      if(error) then
          open(unit=2,file='tt.dat')
          do ir=1,nr
              write(2,*) r(ir),temp(ir),jj(ir),hh(ir),fj(ir),rk(ir)
          enddo
          close(1)
          stop 1898
      endif
c     
      end







c     --------------------------------------------------------------
c           COMPUTE THE DUST TEMPERATURE FROM MEAN INTENSITY
c
c     This routine will solve for all the temperatures, given the
c     mean intensity, the J-mean opacity, and an external source
c     qfpi.
c
c     ARGUMENTS:
c        qfpi         Source term in erg/s/g/ster
c        jj           Mean intensity in erg/s/cm^3/ster
c        kapj         J-mean opacity
c
c     OUTPUT:
c        temp         Dust temperature
c
c     --------------------------------------------------------------
      subroutine compute_dust_temp_from_jj(nr,qfpi,jj,kapj,temp)
      implicit none
      integer nr
      doubleprecision qfpi(FRSIZE_R),jj(FRSIZE_R),kapj(FRSIZE_R)
      doubleprecision temp(FRSIZE_R),kappa,ti,tiold,find_planckopac
c
      integer ir,i,irr
c
c     Loop over space
c
      do ir=1,nr
c
c         Make a first guess for the temperature
c
          ti = 15.3421d0 * ( jj(ir) + qfpi(ir)/kapj(ir) )**0.25
c
c         Now do a convergence loop for the temperature
c
          do i=1,60
              kappa = find_planckopac(ti,ir)
              tiold = ti
              ti    = 15.3421d0 * ( ( jj(ir)*kapj(ir) + qfpi(ir) ) 
     %                  / kappa )**0.25
              if(abs(ti-tiold)/(ti+tiold).lt.TEMPACCUR) goto 10
          enddo
          write(*,*) 'No convergence in temperature'
          write(*,*) ir,jj(ir),kapj(ir)
          open(unit=2,file='ttt.dat')
          do irr=1,nr
              write(2,*) irr,jj(irr),qfpi(irr),kapj(irr)
          enddo
          close(2)
          stop 733
 10       continue
          temp(ir) = ti
c
      enddo
c
      end


c     --------------------------------------------------------------
c                     COMPUTE THE PLANCK MEAN OPACITY
c     --------------------------------------------------------------
      subroutine compute_planckopac(nf,nt,nr,nu,t0,t1,kappa)
      implicit none
      integer nt,nf,nr
      doubleprecision t0,t1,nu(FRSIZE_FREQ)
      doubleprecision kappa(FRSIZE_FREQ,FRSIZE_R)
c
#     include "common_transfer.h"
c
      doubleprecision enum,denom,bnucur,bnuold,bplanck
      integer it,inu,ir
c
c     Make array of temperatures
c
      dopac_nt = nt
      if(dopac_nt.gt.NR_TEMPERATURE) then
          write(*,*) 'Too many temperatures in requested for planck'
          write(*,*) ' Increase NR_TEMPERATURE and recompile'
          stop 13
      endif
      do it=1,dopac_nt
          dopac_temp(it)  = t0 * (t1/t0)**((it-1.d0)/(nt-1.d0))
      enddo
c
c     Do a loop over radius
c
      do ir=1,nr
c
c         Do a loop over the discretized temperature
c
          do it=1,dopac_nt
              bnucur          = bplanck(dopac_temp(it),nu(1))
              enum            = 0.d0
              denom           = 0.d0
              do inu=2,nf
                  bnuold  = bnucur
                  bnucur  = bplanck(dopac_temp(it),nu(inu))
                  enum    = enum + 0.5 * ( kappa(inu,ir) * bnucur +
     %                      kappa(inu-1,ir) * bnuold ) *
     %                      abs( nu(inu) - nu(inu-1) )
                  denom   = denom + 0.5 * ( bnucur + bnuold ) *
     %                      abs( nu(inu) - nu(inu-1) )
              enddo
              dopac_kappa(it,ir) = enum / denom
          enddo
      enddo
c
      end


c     --------------------------------------------------------------
c                  INTERPOLATE THE PLANCK MEAN OPACITY
c     --------------------------------------------------------------
      function find_planckopac(temp,ir)
      implicit none
      doubleprecision find_planckopac,temp
      integer ir
c
#     include "common_transfer.h"
c
      doubleprecision dum1,dum2,dt
      integer it
c
      if(dopac_nt.le.1) then 
          write(*,*) 'Must prepare planck opacity table first'
          write(*,*) dopac_nt
          stop 13
      endif
c
      if(temp.lt.dopac_temp(1)) then 
          find_planckopac = dopac_kappa(1,ir)
      elseif(temp.gt.dopac_temp(dopac_nt)) then
c          find_planckopac = dopac_kappa(dopac_nt,ir)
          write(*,*) 'Temperature exceeds ',dopac_temp(dopac_nt)
c##########################
          write(*,*) dopac_nt,(dopac_temp(it),it=1,dopac_nt)
c##########################
          write(*,*) 'If you want this to work, then increase'
          write(*,*) 'the upper temperature of planck opac table'
          stop 13
      else
          call hunt(dopac_temp,dopac_nt,temp,it)
          dt = (temp-dopac_temp(it))/
     %            (dopac_temp(it+1)-dopac_temp(it))
          if((dt.lt.0.d0).or.(dt.gt.1.d0)) stop 552
          find_planckopac = (1.d0-dt)*dopac_kappa(it,ir)+
     %       dt*dopac_kappa(it+1,ir)
      endif
c
      return
      end





c     --------------------------------------------------------------
c                CORRECT THE FLUX FOR STABILITY
c
c     This routine applies a small correction to the H_nu flux from
c     the transfer routine, such that the total flux H is equal to 
c     the flux H following from the moment equations. This turns out
c     to be a crucial stabilizing factor for the variable eddington 
c     factor method. 
c
c     ARGUMENTS:
c       nu        Frequency in Hz
c       hnu       Eddington flux in erg/cm^2/s/Hz
c       hh        Eddington flux (freq-integrated) in erg/cm^2/s
c       kappa     Opacity in cm^2/g
c       temp      Temperature in K
c
c     RESULTS:
c       hnu       New (corrected) Eddington flux in erg/cm^2/s/Hz
c       corfact   (for debug) How much the flux has been corrected? 
c
c     --------------------------------------------------------------
      subroutine correct_flux(nf,nu,hnu,hh,alpha,temp,corfact)
      implicit none
      integer nf
      doubleprecision hnu(FRSIZE_FREQ),hh,temp,corfact
      doubleprecision nu(FRSIZE_FREQ),alpha(FRSIZE_FREQ)
c
      doubleprecision htildenu(FRSIZE_FREQ),htilde,bplanckdt
      doubleprecision chi,hnow,h1,h2
      integer inu
c
c     Find out what the total flux is in Hnu
c
      hnow = 0.d0
      do inu=2,nf
          hnow = hnow + 0.5 * ( hnu(inu) + hnu(inu-1) ) *
     %             abs( nu(inu) - nu(inu-1) )
      enddo
c
c     Now make tilde H_nu, i.e. a dummy flux proportional to
c     what a termal flux (1/alphanu)(dBnu/dT)(dT/dz) would be.
c
      do inu=1,nf
          htildenu(inu) = bplanckdt(temp,nu(inu)) / alpha(inu)
      enddo
c
c     Find out the norm of this dummy flux
c
      htilde = 0.d0
      do inu=2,nf
          htilde = htilde + 0.5 * 
     %          ( htildenu(inu) + htildenu(inu-1) ) *
     %            abs( nu(inu) - nu(inu-1) )
      enddo
c
c     Find out how much of this flux is needed to add to Hnu in
c     order to make the freq-integral of the new Hnu equal to
c     the flux from the moment equations
c
      chi = ( hh - hnow ) / htilde
c
c     Now correct the flux to achieve this goal
c
      do inu=1,nf
          hnu(inu) = hnu(inu) + chi * htildenu(inu)
      enddo
c
c     For curiosity, we compute how large this correction was
c     (is not part of the algorithm!)
c
      h1 = 0.d0
      h2 = 0.d0
      do inu=2,nf
          h1 = h1 + 0.5 * abs( htildenu(inu) + htildenu(inu-1) ) *
     %            abs( nu(inu) - nu(inu-1) )
          h2 = h2 + 0.5 * abs( hnu(inu) + hnu(inu-1) ) *
     %            abs( nu(inu) - nu(inu-1) )
      enddo
      corfact = h1*abs(chi)/h2
c
      end


c     --------------------------------------------------------------
c           THE TEMPERATURE DERIVATIVE OF PLANCK FUNCTION 
c
c     This function computes the temperature derivative of the
c     Blackbody function 
c
c        dB_nu(T)     2 h^2 nu^4      exp(h nu / kT)        1 
c        --------   = ---------- ------------------------  ---
c           dT          k c^2    [ exp(h nu / kT) - 1 ]^2  T^2
c
c     ARGUMENTS:
c        nu    [Hz]            = Frequency
c        temp  [K]             = Electron temperature
c     --------------------------------------------------------------
      function bplanckdt(temp,nu)
      implicit none
      doubleprecision temp
      doubleprecision nu
      doubleprecision bplanckdt,theexp
c
#     ifndef NOCHECK
      if(temp.le.0.d0) then
          write(*,*) 'ERROR: bplanckdt()'
          write(*,*) '       temperature = ',temp,' K'
          stop 13
      endif
#     endif
c
      theexp = exp(4.7989d-11*nu/temp)
      if(theexp.lt.1.d33) then
          bplanckdt = 7.07661334104d-58 * nu**4 * 
     %         theexp / ( (theexp-1.d0)**2 * temp**2 )
     %         + 1.d-290
      else
          bplanckdt = 7.07661334104d-58 * nu**4 
     %         / ( theexp * temp**2 )
     %         + 1.d-290
      endif
c     
      return
      end


